
//////////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2007-2018 zSpace, Inc.  All Rights Reserved.
//
//  File:       Sdk4.jaja
//  Content:    Java bindings for the 4.0 zSpace SDK.
//
//////////////////////////////////////////////////////////////////////////

package com.zspace;

public class Sdk4 {

  static final int ZC_API_VERSION = 2;

  //////////////////////////////////////////////////////////////////////////
  // Enums
  //////////////////////////////////////////////////////////////////////////

  /// Defines the error codes returned by all Core SDK functions.
  public enum ZCError
  {
    ZC_ERROR_OK(0),
    ZC_ERROR_NOT_IMPLEMENTED(1),
    ZC_ERROR_NOT_INITIALIZED(2),
    ZC_ERROR_ALREADY_INITIALIZED(3),
    ZC_ERROR_INVALID_PARAMETER(4),
    ZC_ERROR_INVALID_CONTEXT(5),
    ZC_ERROR_INVALID_HANDLE(6),
    ZC_ERROR_RUNTIME_INCOMPATIBLE(7),
    ZC_ERROR_RUNTIME_NOT_FOUND(8),
    ZC_ERROR_SYMBOL_NOT_FOUND(9),
    ZC_ERROR_DISPLAY_NOT_FOUND(10),
    ZC_ERROR_DEVICE_NOT_FOUND(11),
    ZC_ERROR_TARGET_NOT_FOUND (12),
    ZC_ERROR_CAPABILITY_NOT_FOUND(13),
    ZC_ERROR_BUFFER_TOO_SMALL(14),
    ZC_ERROR_SYNC_FAILED(15),
    ZC_ERROR_OPERATION_FAILED(16),
    ZC_ERROR_INVALID_ATTRIBUTE(17);

    public final int Value;

    private ZCError(int value)
    {
        Value = value;
    }

    public static ZCError toEnum(int value) 
    {
      ZCError[] values = ZCError.class.getEnumConstants();
      if (value < values.length && value >= 0 && values[value].Value == value)
      {
        return values[value];
      }

      for (ZCError enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCError.class + " with value " + value);
    }
  }


  /// Defines the types of displays for the Display APIs.
  public enum ZCDisplayType
  {
    ZC_DISPLAY_TYPE_UNKNOWN(-1),
    ZC_DISPLAY_TYPE_GENERIC(0),
    ZC_DISPLAY_TYPE_ZSPACE(1);

    public final int Value;

    private ZCDisplayType(int value)
    {
        Value = value;
    }

    public static ZCDisplayType toEnum(int value) 
    {
      ZCDisplayType[] values = ZCDisplayType.class.getEnumConstants();
      for (ZCDisplayType enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCDisplayType.class + " with value " + value);
    }
  }


  /// Defines the attributes that you can query for the display.
  /// See zcGetDisplayAttributeStr().
  public enum ZCDisplayAttribute
  {
    ZC_DISPLAY_ATTRIBUTE_ADAPTER_NAME(0),         ///< The graphics adapter name.
    ZC_DISPLAY_ATTRIBUTE_ADAPTER_STRING(1),       ///< The graphics adapter context string.
    ZC_DISPLAY_ATTRIBUTE_ADAPTER_ID(2),           ///< The entire ID string of the graphics adapter.
    ZC_DISPLAY_ATTRIBUTE_ADAPTER_VENDOR_ID(3),    ///< The vendor ID of the graphics adapter.
    ZC_DISPLAY_ATTRIBUTE_ADAPTER_DEVICE_ID(4),    ///< The device ID of the graphics adapter.
    ZC_DISPLAY_ATTRIBUTE_ADAPTER_KEY(5),          ///< Reserved.
    ZC_DISPLAY_ATTRIBUTE_MONITOR_NAME(6),         ///< The monitor name.
    ZC_DISPLAY_ATTRIBUTE_MONITOR_STRING(7),       ///< The monitor context string.
    ZC_DISPLAY_ATTRIBUTE_MONITOR_ID(8),           ///< The entire ID string of the monitor.
    ZC_DISPLAY_ATTRIBUTE_MONITOR_VENDOR_ID(9),    ///< The vendor ID of the monitor.
    ZC_DISPLAY_ATTRIBUTE_MONITOR_DEVICE_ID(10),   ///< The device ID of the monitor.
    ZC_DISPLAY_ATTRIBUTE_MONITOR_KEY(11),         ///< Reserved.
    ZC_DISPLAY_ATTRIBUTE_MANUFACTURER_NAME(12),   ///< The display's manufacturer name.
    ZC_DISPLAY_ATTRIBUTE_PRODUCT_CODE(13),        ///< The display's product code.
    ZC_DISPLAY_ATTRIBUTE_SERIAL_NUMBER(14),       ///< The display's serial number.
    ZC_DISPLAY_ATTRIBUTE_VIDEO_INTERFACE(15),     ///< The display's video interface.
    ZC_DISPLAY_ATTRIBUTE_MODEL(16);               ///< The display's hardware model (currently only supported for zSpace displays).

    public final int Value;

    private ZCDisplayAttribute(int value)
    {
        Value = value;
    }

    public static ZCDisplayAttribute toEnum(int value) 
    {
      ZCDisplayAttribute[] values = ZCDisplayAttribute.class.getEnumConstants();
      if (value < values.length && value >= 0 && values[value].Value == value)
      {
        return values[value];
      }

      for (ZCDisplayAttribute enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCDisplayAttribute.class + " with value " + value);
    }
  }


  /// Defines the renderer used by the Stereo Buffer API.
  public enum ZCRenderer
  {
    ZC_RENDERER_QUAD_BUFFER_GL(0),
    ZC_RENDERER_QUAD_BUFFER_D3D9(1),
    ZC_RENDERER_QUAD_BUFFER_D3D10(2),
    ZC_RENDERER_QUAD_BUFFER_D3D11(3),
    ZC_RENDERER_QUAD_BUFFER_D3D11_1(4);

    public final int Value;

    private ZCRenderer(int value)
    {
        Value = value;
    }

    public static ZCRenderer toEnum(int value) 
    {
      ZCRenderer[] values = ZCRenderer.class.getEnumConstants();
      if (value < values.length && value >= 0 && values[value].Value == value)
      {
        return values[value];
      }

      for (ZCRenderer enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCRenderer.class + " with value " + value);
    }
  }


  /// Defines the eyes for the Stereo Frustum API.
  /// This enum is also used by the Stereo Buffer API.
  public enum ZCEye
  {
    ZC_EYE_LEFT(0),
    ZC_EYE_RIGHT(1),
    ZC_EYE_CENTER(2);

    public final int Value;

    private ZCEye(int value)
    {
        Value = value;
    }

    public static ZCEye toEnum(int value) 
    {
      ZCEye[] values = ZCEye.class.getEnumConstants();
      if (value < values.length && value >= 0 && values[value].Value == value)
      {
        return values[value];
      }

      for (ZCEye enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCEye.class + " with value " + value);
    }
  }


  /// Defines the coordinate spaces used by the zSpace Core SDK.
  /// This enum is used by both the Coordinate Space API and 
  /// the Stereo Frustum API.
  public enum ZCCoordinateSpace
  {
    ZC_COORDINATE_SPACE_TRACKER(0),
    ZC_COORDINATE_SPACE_DISPLAY(1),
    ZC_COORDINATE_SPACE_VIEWPORT(2),
    ZC_COORDINATE_SPACE_CAMERA(3);

    public final int Value;

    private ZCCoordinateSpace(int value)
    {
        Value = value;
    }

    public static ZCCoordinateSpace toEnum(int value) 
    {
      ZCCoordinateSpace[] values = ZCCoordinateSpace.class.getEnumConstants();
      if (value < values.length && value >= 0 && values[value].Value == value)
      {
        return values[value];
      }

      for (ZCCoordinateSpace enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCCoordinateSpace.class + " with value " + value);
    }
  }


  /// Defines the attributes that you can set and query for the StereoFrustum.
  /// These attributes are important for comfortable viewing of stereoscopic 3D.
  public enum ZCFrustumAttribute
  {
    ///////////////////////////////////////////
    // Float Attributes
    ///////////////////////////////////////////

    /// The physical separation, or inter-pupillary distance, between the eyes in meters.
    /// An IPD of 0 will effectively disable stereo since the eyes are assumed
    /// to be at the same location. (Default: 0.060)
    ZC_FRUSTUM_ATTRIBUTE_IPD(0x00000000),

    /// Viewer scale adjusts the display and head tracking for larger and smaller scenes. (Default: 1)
    /// Use larger values for scenes with large models and smaller values for smaller models.
    ZC_FRUSTUM_ATTRIBUTE_VIEWER_SCALE(0x00000001),

    /// Uniform scale factor applied to the frustum's incoming head pose. (Default: 1)
    ZC_FRUSTUM_ATTRIBUTE_HEAD_SCALE(0x00000003),

    /// Near clipping plane for the frustum in meters. (Default: 0.1)
    ZC_FRUSTUM_ATTRIBUTE_NEAR_CLIP(0x00000004),

    /// Far clipping plane for the frustum in meters. (Default: 1000)
    ZC_FRUSTUM_ATTRIBUTE_FAR_CLIP(0x00000005),

    /// Distance between the bridge of the glasses and the bridge of the nose in meters. (Default: 0.01)
    ZC_FRUSTUM_ATTRIBUTE_GLASSES_OFFSET(0x00000006),

    /// Maximum pixel disparity for crossed images (negative parallax) in the coupled zone. (Default: -100)
    /// The coupled zone refers to the area where our eyes can both comfortably converge and focus on an object. 
    ZC_FRUSTUM_ATTRIBUTE_CC_LIMIT(0x00000007),

    /// Maximum pixel disparity for uncrossed images (positive parallax) in the coupled zone. (Default: 100)
    ZC_FRUSTUM_ATTRIBUTE_UC_LIMIT(0x00000008),

    /// Maximum pixel disparity for crossed images (negative parallax) in the uncoupled zone. (Default: -200)
    ZC_FRUSTUM_ATTRIBUTE_CU_LIMIT(0x00000009),

    /// Maximum pixel disparity for uncrossed images (positive parallax) in the uncoupled zone. (Default: 250)
    ZC_FRUSTUM_ATTRIBUTE_UU_LIMIT(0x0000000A),

    /// Maximum depth in meters for negative parallax in the coupled zone. (Default: 0.13)
    ZC_FRUSTUM_ATTRIBUTE_CC_DEPTH(0x0000000B),

    /// Maximum depth in meters for positive parallax in the coupled zone. (Default: -0.30)
    ZC_FRUSTUM_ATTRIBUTE_UC_DEPTH(0x0000000C),

    /// Display angle in degrees about the X axis. (Default: 30.0)
    /// Is only used when ZC_PORTAL_MODE_ANGLE is not enabled on the frustum.
    ZC_FRUSTUM_ATTRIBUTE_DISPLAY_ANGLE_X(0x0000000D),

    /// Display angle in degrees about the Y axis. (Default: 0.0)
    /// Is only used when ZC_PORTAL_MODE_ANGLE is not enabled on the frustum.
    ZC_FRUSTUM_ATTRIBUTE_DISPLAY_ANGLE_Y(0x0000000E),

    /// Display angle in degrees about the Z axis. (Default: 0.0)
    /// Is only used when ZC_PORTAL_MODE_ANGLE is not enabled on the frustum.
    ZC_FRUSTUM_ATTRIBUTE_DISPLAY_ANGLE_Z(0x0000000F),

    /// The delay in seconds before the automatic transition from stereo to mono begins. (Default: 5.0)
    ZC_FRUSTUM_ATTRIBUTE_AUTO_STEREO_DELAY(0x00000010),

    /// The duration in seconds of the automatic transition from stereo to mono. (Default: 1.0)
    ZC_FRUSTUM_ATTRIBUTE_AUTO_STEREO_DURATION(0x00000011),


    ///////////////////////////////////////////
    // Boolean Attributes
    ///////////////////////////////////////////

    /// Flag controlling whether the automatic transition from stereo to mono is enabled. (Default: true)
    ZC_FRUSTUM_ATTRIBUTE_AUTO_STEREO_ENABLED (0x00010000);

    public final int Value;

    private ZCFrustumAttribute(int value)
    {
        Value = value;
    }

    public static ZCFrustumAttribute toEnum(int value) 
    {
      ZCFrustumAttribute[] values = ZCFrustumAttribute.class.getEnumConstants();
      for (ZCFrustumAttribute enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCFrustumAttribute.class + " with value " + value);
    }
  }

  /// Defines options for positioning the scene relative to the physical display or relative to the viewport.
  static final int ZC_PORTAL_MODE_NONE = 0x00000000;
  static final int ZC_PORTAL_MODE_ANGLE = 0x00000001;
  static final int ZC_PORTAL_MODE_POSITION = 0x00000002;
  static final int ZC_PORTAL_MODE_ALL = 0xFFFFFFFF;


  /// Defines the types of tracker targets.
  public enum ZCTargetType
  {
    ZC_TARGET_TYPE_HEAD(0),       ///< The tracker target corresponding to the user's head.
    ZC_TARGET_TYPE_PRIMARY(1),    ///< The tracker target corresponding to the user's primary hand.
    ZC_TARGET_TYPE_SECONDARY(2);  ///< The tracker target corresponding to the user's secondary hand. (Reserved for future use.)

    public final int Value;

    private ZCTargetType(int value)
    {
        Value = value;
    }

    public static ZCTargetType toEnum(int value) 
    {
      ZCTargetType[] values = ZCTargetType.class.getEnumConstants();
      if (value < values.length && value >= 0 && values[value].Value == value)
      {
        return values[value];
      }

      for (ZCTargetType enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCTargetType.class + " with value " + value);
    }
  }


  /// Defines event types that you can use with a tracker event handler.
  public enum ZCTrackerEventType
  {
    // Target Events
    ZC_TRACKER_EVENT_MOVE(0x0101),

    // Button Events
    ZC_TRACKER_EVENT_BUTTON_PRESS(0x0201),
    ZC_TRACKER_EVENT_BUTTON_RELEASE(0x0202),

    // Tap Events
    ZC_TRACKER_EVENT_TAP_PRESS(0x0301),
    ZC_TRACKER_EVENT_TAP_RELEASE(0x0302),
    ZC_TRACKER_EVENT_TAP_HOLD(0x0303),
    ZC_TRACKER_EVENT_TAP_SINGLE(0x0304),
    ZC_TRACKER_EVENT_TAP_DOUBLE(0x0305),

    // All Events
    ZC_TRACKER_EVENT_ALL(0xFFFF);

    public final int Value;

    private ZCTrackerEventType(int value)
    {
        Value = value;
    }

    public static ZCTrackerEventType toEnum(int value) 
    {
      ZCTrackerEventType[] values = ZCTrackerEventType.class.getEnumConstants();
      for (ZCTrackerEventType enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCTrackerEventType.class + " with value " + value);
    }
  }


  /// Defines mouse buttons to be used when mapping a tracker target's buttons to a mouse.
  public enum ZCMouseButton
  {
    ZC_MOUSE_BUTTON_UNKNOWN(-1),
    ZC_MOUSE_BUTTON_LEFT(0),
    ZC_MOUSE_BUTTON_RIGHT(1),
    ZC_MOUSE_BUTTON_CENTER(2);

    public final int Value;

    private ZCMouseButton(int value)
    {
        Value = value;
    }

    public static ZCMouseButton toEnum(int value) 
    {
      ZCMouseButton[] values = ZCMouseButton.class.getEnumConstants();
      for (ZCMouseButton enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCMouseButton.class + " with value " + value);
    }
  }


  /// Determines how the stylus and mouse control the cursor when both are used.
  public enum ZCMouseMovementMode
  {
    /// The stylus uses absolute positions.  
    /// In this mode, the mouse and stylus can fight for control of the cursor if both are in use.
    /// This is the default mode. 
    ZC_MOUSE_MOVEMENT_MODE_ABSOLUTE(0),

    /// The stylus applies delta positions to the mouse cursor's current position.
    /// Movements by the mouse and stylus are compounded without fighting. 
    ZC_MOUSE_MOVEMENT_MODE_RELATIVE(1);

    public final int Value;

    private ZCMouseMovementMode(int value)
    {
        Value = value;
    }

    public static ZCMouseMovementMode toEnum(int value) 
    {
      ZCMouseMovementMode[] values = ZCMouseMovementMode.class.getEnumConstants();
      for (ZCMouseMovementMode enumValue : values)
      {
        if (enumValue.Value == value)
        {
          return enumValue;
        }
      }
    
      throw new IllegalArgumentException("No enum " + ZCMouseMovementMode.class + " with value " + value);
    }
  }

  //////////////////////////////////////////////////////////////////////////
  // zSpace APIs
  //////////////////////////////////////////////////////////////////////////

  // General API
  public static long zcInitialize()
  {
    return Sdk4JNI.zcInitialize();
  }

  public static void zcUpdate(long context)
  {
    Sdk4JNI.zcUpdate(context);
    dispatchEvents();
  }

  public static void zcShutDown(long context)
  {
    Sdk4JNI.zcShutDown(context);
  }

  public static long zcGetContext(long handle)
  {
    return Sdk4JNI.zcGetContext(handle);
  }

  public static void zcSetTrackingEnabled(long context, boolean isEnabled)
  {
    Sdk4JNI.zcSetTrackingEnabled(context, isEnabled);
  }

  public static boolean zcIsTrackingEnabled(long context)
  {
    return Sdk4JNI.zcIsTrackingEnabled(context);
  }

  public static void zcGetRuntimeVersion(long context, ZCRuntimeVersion version)
  {
    Sdk4JNI.zcGetRuntimeVersion(context, version);
  }

  public static ZCError zcGetError()
  {
    return ZCError.toEnum(Sdk4JNI.zcGetError());
  }

  public static void zcClearError()
  {
    Sdk4JNI.zcClearError();
  }

  public static String zcGetErrorString(Sdk4.ZCError error)
  {
    return Sdk4JNI.zcGetErrorString(error.Value);
  }

  // Display API
  public static void zcRefreshDisplays(long context)
  {
    Sdk4JNI.zcRefreshDisplays(context);
  }

  public static int zcGetNumDisplays(long context)
  {
    return Sdk4JNI.zcGetNumDisplays(context);
  }

  public static int zcGetNumDisplaysByType(long context, Sdk4.ZCDisplayType displayType)
  {
    return Sdk4JNI.zcGetNumDisplaysByType(context, displayType.Value);
  }

  public static long zcGetDisplay(long context, int x, int y)
  {
    return Sdk4JNI.zcGetDisplay(context, x, y);
  }

  public static long zcGetDisplayByIndex(long context, int index)
  {
    return Sdk4JNI.zcGetDisplayByIndex(context, index);
  }

  public static long zcGetDisplayByType(long context, Sdk4.ZCDisplayType displayType, int index)
  {
    return Sdk4JNI.zcGetDisplayByType(context, displayType.Value, index);
  }

  public static Sdk4.ZCDisplayType zcGetDisplayType(long displayHandle)
  {
    int displayType = Sdk4JNI.zcGetDisplayType(displayHandle);;
    return Sdk4.ZCDisplayType.toEnum(displayType);
  }

  public static int zcGetDisplayNumber(long displayHandle)
  {
    return Sdk4JNI.zcGetDisplayNumber(displayHandle);
  }

  public static int zcGetDisplayAdapterIndex(long displayHandle)
  {
    return Sdk4JNI.zcGetDisplayAdapterIndex(displayHandle);
  }

  public static int zcGetDisplayMonitorIndex(long displayHandle)
  {
    return Sdk4JNI.zcGetDisplayMonitorIndex(displayHandle);
  }

  public static String zcGetDisplayAttributeStr(long displayHandle, Sdk4.ZCDisplayAttribute attribute)
  {
    return Sdk4JNI.zcGetDisplayAttributeStr(displayHandle, attribute.Value);
  }

  public static void zcGetDisplaySize(long displayHandle, ZSVector2 size)
  {
    Sdk4JNI.zcGetDisplaySize(displayHandle, size);
  }

  public static void zcGetDisplayPosition(long displayHandle, ZSVector2 position)
  {
    Sdk4JNI.zcGetDisplayPosition(displayHandle, position);
  }

  public static void zcGetDisplayNativeResolution(long displayHandle, ZSVector2 resolution)
  {
    Sdk4JNI.zcGetDisplayNativeResolution(displayHandle, resolution);
  }

  public static void zcGetDisplayAngle(long displayHandle, ZSVector3 angle)
  {
    Sdk4JNI.zcGetDisplayAngle(displayHandle, angle);
  }

  public static float zcGetDisplayVerticalRefreshRate(long displayHandle)
  {
    return Sdk4JNI.zcGetDisplayVerticalRefreshRate(displayHandle);
  }

  public static boolean zcIsDisplayHardwarePresent(long displayHandle)
  {
    return Sdk4JNI.zcIsDisplayHardwarePresent(displayHandle);
  }

  public static void zcIntersectDisplay(long displayHandle, ZCTrackerPose pose, ZCDisplayIntersectionInfo intersectionInfo)
  {
    Sdk4JNI.zcIntersectDisplay(displayHandle, pose, intersectionInfo);
  }

  // StereoBuffer API
  public static long zcCreateStereoBuffer(long context, Sdk4.ZCRenderer renderer, long reserved)
  {
    return Sdk4JNI.zcCreateStereoBuffer(context, renderer.Value, reserved);
  }

  public static void zcDestroyStereoBuffer(long bufferHandle)
  {
    Sdk4JNI.zcDestroyStereoBuffer(bufferHandle);
  }

  public static void zcSetStereoBufferFullScreen(long bufferHandle, boolean isFullScreen)
  {
    Sdk4JNI.zcSetStereoBufferFullScreen(bufferHandle, isFullScreen);
  }

  public static boolean zcIsStereoBufferFullScreen(long bufferHandle)
  {
    return  Sdk4JNI.zcIsStereoBufferFullScreen(bufferHandle);
  }

  public static void zcBeginStereoBufferFrame(long bufferHandle)
  {
    Sdk4JNI.zcBeginStereoBufferFrame(bufferHandle);
  }

  public static void zcSyncStereoBuffer(long bufferHandle)
  {
    Sdk4JNI.zcSyncStereoBuffer(bufferHandle);
  }

  // StereoViewport API
  public static long zcCreateViewport(long context)
  {
    return Sdk4JNI.zcCreateViewport(context);
  }

  public static void zcDestroyViewport(long viewportHandle)
  {
    Sdk4JNI.zcDestroyViewport(viewportHandle);
  }

  public static void zcSetViewportPosition(long viewportHandle, int x, int y)
  {
    Sdk4JNI.zcSetViewportPosition(viewportHandle, x, y);
  }

  public static void zcGetViewportPosition(long viewportHandle, ZSVector2 position)
  {
    Sdk4JNI.zcGetViewportPosition(viewportHandle, position);
  }

  public static void zcSetViewportSize(long viewportHandle, int width, int height)
  {
    Sdk4JNI.zcSetViewportSize(viewportHandle, width, height);
  }

  public static void zcGetViewportSize(long viewportHandle, ZSVector2 size)
  {
    Sdk4JNI.zcGetViewportSize(viewportHandle, size);
  }

  // Coordinate Space API
  public static void zcGetCoordinateSpaceTransform(long viewportHandle, Sdk4.ZCCoordinateSpace a, Sdk4.ZCCoordinateSpace b, ZSMatrix4 transform)
  {
    Sdk4JNI.zcGetCoordinateSpaceTransform(viewportHandle, a.Value, b.Value, transform.f);
  }

  public static void zcTransformMatrix(long viewportHandle, Sdk4.ZCCoordinateSpace a, Sdk4.ZCCoordinateSpace b, ZSMatrix4 matrix)
  {
    Sdk4JNI.zcTransformMatrix(viewportHandle, a.Value, b.Value, matrix.f);
  }

  // StereoFrustum API
  public static long zcGetFrustum(long viewportHandle)
  {
    return Sdk4JNI.zcGetFrustum(viewportHandle);
  }

  public static void zcSetFrustumAttributeF32(long frustumHandle, Sdk4.ZCFrustumAttribute attribute, float value)
  {
    Sdk4JNI.zcSetFrustumAttributeF32(frustumHandle, attribute.Value, value);
  }

  public static float zcGetFrustumAttributeF32(long frustumHandle, Sdk4.ZCFrustumAttribute attribute)
  {
    return Sdk4JNI.zcGetFrustumAttributeF32(frustumHandle, attribute.Value);
  }

  public static void zcSetFrustumAttributeB(long frustumHandle, Sdk4.ZCFrustumAttribute attribute, boolean value)
  {
    Sdk4JNI.zcSetFrustumAttributeB(frustumHandle, attribute.Value, value);
  }

  public static boolean zcGetFrustumAttributeB(long frustumHandle, Sdk4.ZCFrustumAttribute attribute)
  {
    return Sdk4JNI.zcGetFrustumAttributeB(frustumHandle, attribute.Value);
  }

  public static void zcSetFrustumPortalMode(long frustumHandle, int portalModeFlags)
  {
    Sdk4JNI.zcSetFrustumPortalMode(frustumHandle, portalModeFlags);
  }

  public static int zcGetFrustumPortalMode(long frustumHandle)
  {
    return Sdk4JNI.zcGetFrustumPortalMode(frustumHandle);
  }

  public static void zcSetFrustumCameraOffset(long frustumHandle, ZSVector3 cameraOffset)
  {
    Sdk4JNI.zcSetFrustumCameraOffset(frustumHandle, cameraOffset.x, cameraOffset.y, cameraOffset.z);
  }

  public static void zcGetFrustumCameraOffset(long frustumHandle, ZSVector3 cameraOffset)
  {
    Sdk4JNI.zcGetFrustumCameraOffset(frustumHandle, cameraOffset);
  }

  public static void zcSetFrustumHeadPose(long frustumHandle, ZCTrackerPose headPose)
  {
    Sdk4JNI.zcSetFrustumHeadPose(frustumHandle, headPose.timestamp, headPose.matrix.f);
  }

  public static void zcGetFrustumHeadPose(long frustumHandle, ZCTrackerPose headPose)
  {
     headPose.timestamp = Sdk4JNI.zcGetFrustumHeadPose(frustumHandle, headPose.matrix.f);
  }

  public static void zcGetFrustumViewMatrix(long frustumHandle, Sdk4.ZCEye eye, ZSMatrix4 viewMatrix)
  {
    Sdk4JNI.zcGetFrustumViewMatrix(frustumHandle, eye.Value, viewMatrix.f);
  }

  public static void zcGetFrustumProjectionMatrix(long frustumHandle, Sdk4.ZCEye eye, ZSMatrix4 projectionMatrix)
  {
    Sdk4JNI.zcGetFrustumProjectionMatrix(frustumHandle, eye.Value, projectionMatrix.f);
  }

  public static void zcGetFrustumBounds(long frustumHandle, Sdk4.ZCEye eye, ZCFrustumBounds bounds)
  {
    Sdk4JNI.zcGetFrustumBounds(frustumHandle, eye.Value, bounds);
  }

  public static void zcGetFrustumEyePosition(long frustumHandle, Sdk4.ZCEye eye, Sdk4.ZCCoordinateSpace coordinateSpace, ZSVector3 eyePosition)
  {
    Sdk4JNI. zcGetFrustumEyePosition(frustumHandle, eye.Value, coordinateSpace.Value, eyePosition);
  }

  public static void zcGetFrustumCoupledBoundingBox(long frustumHandle, ZCBoundingBox boundingBox)
  {
    Sdk4JNI.zcGetFrustumCoupledBoundingBox(frustumHandle, boundingBox.lower, boundingBox.upper);
  }

  public static float zcCalculateFrustumFit(long frustumHandle, ZCBoundingBox boundingBox, ZSMatrix4 lookAtMatrix)
  {
    return Sdk4JNI.zcCalculateFrustumFit(frustumHandle, boundingBox.lower.x, boundingBox.lower.y, boundingBox.lower.z, boundingBox.upper.x, boundingBox.upper.y, boundingBox.upper.z, lookAtMatrix.f);
  }

  public static float zcCalculateFrustumDisparity(long frustumHandle, ZSVector3 point)
  {
    return Sdk4JNI.zcCalculateFrustumDisparity(frustumHandle, point.x, point.y, point.z);
  }

  // TrackerDevice API
  public static int zcGetNumTrackerDevices(long context)
  {
    return Sdk4JNI.zcGetNumTrackerDevices(context);
  }

  public static long zcGetTrackerDevice(long context, int index)
  {
    return Sdk4JNI.zcGetTrackerDevice(context, index);
  }

  public static long zcGetTrackerDeviceByName(long context, String deviceName)
  {
    return Sdk4JNI.zcGetTrackerDeviceByName(context, deviceName);
  }

  public static void zcSetTrackerDeviceEnabled(long deviceHandle, boolean isEnabled)
  {
    Sdk4JNI.zcSetTrackerDeviceEnabled(deviceHandle, isEnabled);
  }

  public static boolean zcIsTrackerDeviceEnabled(long deviceHandle)
  {
    return Sdk4JNI.zcIsTrackerDeviceEnabled(deviceHandle);
  }

  public static String zcGetTrackerDeviceName(long deviceHandle)
  {
    return Sdk4JNI.zcGetTrackerDeviceName(deviceHandle);
  }


  // TrackerTarget API
  public static int zcGetNumTargets(long deviceHandle)
  {
    return Sdk4JNI.zcGetNumTargets(deviceHandle);
  }

  public static int zcGetNumTargetsByType(long context, Sdk4.ZCTargetType targetType)
  {
    return Sdk4JNI.zcGetNumTargetsByType(context, targetType.Value);
  }

  public static long zcGetTarget(long deviceHandle, int index)
  {
    return Sdk4JNI.zcGetTarget(deviceHandle, index);
  }

  public static long zcGetTargetByName(long deviceHandle, String targetName)
  {
    return Sdk4JNI.zcGetTargetByName(deviceHandle, targetName);
  }

  public static long zcGetTargetByType(long context, Sdk4.ZCTargetType targetType, int index)
  {
    return Sdk4JNI.zcGetTargetByType(context, targetType.Value, index);
  }

  public static String zcGetTargetName(long targetHandle)
  {
    return Sdk4JNI.zcGetTargetName(targetHandle);
  }

  public static void zcSetTargetEnabled(long targetHandle, boolean isEnabled)
  {
    Sdk4JNI.zcSetTargetEnabled(targetHandle, isEnabled);
  }

  public static boolean zcIsTargetEnabled(long targetHandle)
  {
    return Sdk4JNI.zcIsTargetEnabled(targetHandle);
  }

  public static boolean zcIsTargetVisible(long targetHandle)
  {
    return Sdk4JNI.zcIsTargetVisible(targetHandle);
  }

  public static void zcSetTargetMoveEventThresholds(long targetHandle, float time, float distance, float angle)
  {
    Sdk4JNI.zcSetTargetMoveEventThresholds(targetHandle, time, distance, angle);
  }

  public static void zcGetTargetMoveEventThresholds(long targetHandle, ZSVector3 thresholds)
  {
    Sdk4JNI.zcGetTargetMoveEventThresholds(targetHandle, thresholds);
  }

  public static void zcGetTargetPose(long targetHandle, ZCTrackerPose pose)
  {
    pose.timestamp = Sdk4JNI.zcGetTargetPose(targetHandle, pose.matrix.f);
  }

  public static void zcGetTargetTransformedPose(long targetHandle, long viewportHandle, Sdk4.ZCCoordinateSpace coordinateSpace, ZCTrackerPose pose)
  {
    pose.timestamp = Sdk4JNI.zcGetTargetTransformedPose(targetHandle, viewportHandle, coordinateSpace.Value, pose.matrix.f);
  }

  public static void zcSetTargetPoseBufferingEnabled(long targetHandle, boolean isPoseBufferingEnabled)
  {
    Sdk4JNI.zcSetTargetPoseBufferingEnabled(targetHandle, isPoseBufferingEnabled);
  }

  public static boolean zcIsTargetPoseBufferingEnabled(long targetHandle)
  {
    return Sdk4JNI.zcIsTargetPoseBufferingEnabled(targetHandle);
  }

  public static int zcGetTargetPoseBuffer(long targetHandle, float minDelta, float maxDelta, ZCTrackerPose[] buffer)
  {
    return Sdk4JNI.zcGetTargetPoseBuffer(targetHandle, minDelta, maxDelta, buffer);
  }

  public static void zcResizeTargetPoseBuffer(long targetHandle, int capacity)
  {
    Sdk4JNI.zcResizeTargetPoseBuffer(targetHandle, capacity);
  }

  public static int zcGetTargetPoseBufferCapacity(long targetHandle)
  {
    return Sdk4JNI.zcGetTargetPoseBufferCapacity(targetHandle);
  }

  // TrackerTarget Button API
  public static int zcGetNumTargetButtons(long targetHandle)
  {
    return Sdk4JNI.zcGetNumTargetButtons(targetHandle);
  }

  public static boolean zcIsTargetButtonPressed(long targetHandle, int buttonId)
  {
    return Sdk4JNI.zcIsTargetButtonPressed(targetHandle, buttonId);
  }

  // TrackerTarget Led API
  public static void zcSetTargetLedEnabled(long targetHandle, boolean isLedEnabled)
  {
    Sdk4JNI.zcSetTargetLedEnabled(targetHandle, isLedEnabled);
  }

  public static boolean zcIsTargetLedEnabled(long targetHandle)
  {
    return Sdk4JNI.zcIsTargetLedEnabled(targetHandle);
  }

  public static boolean zcIsTargetLedOn(long targetHandle)
  {
    return Sdk4JNI.zcIsTargetLedOn(targetHandle);
  }

  public static void zcSetTargetLedColor(long targetHandle, float r, float g, float b)
  {
    Sdk4JNI.zcSetTargetLedColor(targetHandle, r, g, b);
  }

  public static void zcGetTargetLedColor(long targetHandle, ZSVector3 color)
  {
    Sdk4JNI.zcGetTargetLedColor(targetHandle, color);
  }

  // TrackerTarget Vibration API
  public static void zcSetTargetVibrationEnabled(long targetHandle, boolean isVibrationEnabled)
  {
    Sdk4JNI.zcSetTargetVibrationEnabled(targetHandle, isVibrationEnabled);
  }

  public static boolean zcIsTargetVibrationEnabled(long targetHandle)
  {
    return Sdk4JNI.zcIsTargetVibrationEnabled(targetHandle);
  }

  public static boolean zcIsTargetVibrating(long targetHandle)
  {
    return Sdk4JNI.zcIsTargetVibrating(targetHandle);
  }

  public static void zcStartTargetVibration(long targetHandle, float onPeriod, float offPeriod, int numTimes, float intensity)
  {
    Sdk4JNI.zcStartTargetVibration(targetHandle, onPeriod, offPeriod, numTimes, intensity);
  }

  public static void zcStopTargetVibration(long targetHandle)
  {
    Sdk4JNI.zcStopTargetVibration(targetHandle);
  }

  // TrackerTarget Tap API
  public static boolean zcIsTargetTapPressed(long targetHandle)
  {
    return Sdk4JNI.zcIsTargetTapPressed(targetHandle);
  }

  public static void zcSetTargetTapHoldThreshold(long targetHandle, float seconds)
  {
    Sdk4JNI.zcSetTargetTapHoldThreshold(targetHandle, seconds);
  }

  public static float zcGetTargetTapHoldThreshold(long targetHandle)
  {
    return Sdk4JNI.zcGetTargetTapHoldThreshold(targetHandle);
  }

  static java.util.ArrayList<ZCEventListener> eventListeners = new java.util.ArrayList<ZCEventListener>();
  static java.util.ArrayList<ZCTrackerEventData> events = new java.util.ArrayList<ZCTrackerEventData>();

  // Tracker Event API
  public static void zcAddTrackerEventHandler(ZCEventListener trackerEventListener)
  {
    eventListeners.add(trackerEventListener);
    Sdk4JNI.zcAddTrackerEventHandler(trackerEventListener.targetHandle, trackerEventListener.trackerEventType.Value, trackerEventListener, trackerEventListener.userData);
  }

  public static void zcRemoveTrackerEventHandler(ZCEventListener trackerEventListener)
  {
    Sdk4JNI.zcRemoveTrackerEventHandler(trackerEventListener.targetHandle, trackerEventListener.trackerEventType.Value, trackerEventListener, trackerEventListener.userData);
    int i=0;
    for (; i<eventListeners.size(); i++)
    {
      ZCEventListener el = eventListeners.get(i);
      if (el.targetHandle == trackerEventListener.targetHandle && 
          el.trackerEventType == trackerEventListener.trackerEventType &&
          el.userData == trackerEventListener.userData)
          break;
    }
    if (i < eventListeners.size())
    {
      eventListeners.remove(i);
    }
  }

  public static void dispatchEvents()
  {
    // Fill in the Events
    Sdk4JNI.getEvents(events);

    // Dispatch them
    for (int i=0; i<events.size(); ++i)
    {
      ZCTrackerEventData event = events.get(i);
      for (int j=0; j<eventListeners.size(); ++j)
      {
        ZCEventListener el = eventListeners.get(j);
        if (el.targetHandle == event.targetHandle && el.trackerEventType.Value == event.type)
        {
          el.runWithEventData(el.targetHandle, event, el.userData);
        }
      }
    }
    events.clear();
  }

  // Mouse Emulation API
  public static void zcSetMouseEmulationEnabled(long context, boolean isEnabled)
  {
    Sdk4JNI.zcSetMouseEmulationEnabled(context, isEnabled);
  }

  public static boolean zcIsMouseEmulationEnabled(long context)
  {
    return Sdk4JNI.zcIsMouseEmulationEnabled(context);
  }

  public static void zcSetMouseEmulationTarget(long context, long targetHandle)
  {
    Sdk4JNI.zcSetMouseEmulationTarget(context, targetHandle);
  }

  public static long zcGetMouseEmulationTarget(long context)
  {
    return Sdk4JNI.zcGetMouseEmulationTarget(context);
  }

  public static void zcSetMouseEmulationMovementMode(long context, Sdk4.ZCMouseMovementMode movementMode)
  {
    Sdk4JNI.zcSetMouseEmulationMovementMode(context, movementMode.Value);
  }

  public static Sdk4.ZCMouseMovementMode zcGetMouseEmulationMovementMode(long context)
  {
    int mode = Sdk4JNI.zcGetMouseEmulationMovementMode(context);
    return Sdk4.ZCMouseMovementMode.toEnum(mode);
  }

  public static void  zcSetMouseEmulationMaxDistance(long context, float maxDistance)
  {
    Sdk4JNI.zcSetMouseEmulationMaxDistance(context, maxDistance);
  }

  public static float zcGetMouseEmulationMaxDistance(long context)
  {
    return Sdk4JNI.zcGetMouseEmulationMaxDistance(context);
  }

  public static void zcSetMouseEmulationButtonMapping(long context, int buttonId, Sdk4.ZCMouseButton mouseButton)
  {
    Sdk4JNI.zcSetMouseEmulationButtonMapping(context, buttonId, mouseButton.Value);
  }

  public static Sdk4.ZCMouseButton zcGetMouseEmulationButtonMapping(long context, int buttonId)
  {
    int button = Sdk4JNI.zcGetMouseEmulationButtonMapping(context, buttonId);
    return Sdk4.ZCMouseButton.toEnum(button);
  }
}
